@@ -61,26 +61,35 @@ namespace dbw_mkz_can
 
 // Latest firmware versions
 PlatformMap FIRMWARE_LATEST({
-  {PlatformVersion(P_FORD_CD4, M_BPEC,  ModuleVersion(2,3,0))},
-  {PlatformVersion(P_FORD_CD4, M_TPEC,  ModuleVersion(2,3,0))},
-  {PlatformVersion(P_FORD_CD4, M_STEER, ModuleVersion(2,3,0))},
-  {PlatformVersion(P_FORD_CD4, M_SHIFT, ModuleVersion(2,3,0))},
-  {PlatformVersion(P_FORD_P5,  M_TPEC,  ModuleVersion(1,2,0))},
-  {PlatformVersion(P_FORD_P5,  M_STEER, ModuleVersion(1,2,0))},
-  {PlatformVersion(P_FORD_P5,  M_SHIFT, ModuleVersion(1,2,0))},
-  {PlatformVersion(P_FORD_P5,  M_ABS,   ModuleVersion(1,2,0))},
-  {PlatformVersion(P_FORD_P5,  M_BOO,   ModuleVersion(1,2,0))},
-  {PlatformVersion(P_FORD_U6,  M_TPEC,  ModuleVersion(0,0,3))},
-  {PlatformVersion(P_FORD_U6,  M_STEER, ModuleVersion(0,0,3))},
-  {PlatformVersion(P_FORD_U6,  M_SHIFT, ModuleVersion(0,0,3))},
-  {PlatformVersion(P_FORD_U6,  M_ABS,   ModuleVersion(0,0,3))},
-  {PlatformVersion(P_FORD_U6,  M_BOO,   ModuleVersion(0,0,3))},
-  {PlatformVersion(P_FORD_C1,  M_TPEC,  ModuleVersion(0,2,0))},
-  {PlatformVersion(P_FORD_C1,  M_STEER, ModuleVersion(0,2,0))},
-  {PlatformVersion(P_FORD_C1,  M_SHIFT, ModuleVersion(0,2,0))},
-  {PlatformVersion(P_FORD_C1,  M_ABS,   ModuleVersion(0,2,0))},
-  {PlatformVersion(P_FORD_C1,  M_BOO,   ModuleVersion(0,2,0))},
-  {PlatformVersion(P_FORD_C1,  M_EPS,   ModuleVersion(0,2,0))},
+  {PlatformVersion(P_FORD_C1,  M_TPEC,  ModuleVersion(1,3,0))},
+  {PlatformVersion(P_FORD_C1,  M_STEER, ModuleVersion(1,3,0))},
+  {PlatformVersion(P_FORD_C1,  M_SHIFT, ModuleVersion(1,3,0))},
+  {PlatformVersion(P_FORD_C1,  M_ABS,   ModuleVersion(1,3,0))},
+  {PlatformVersion(P_FORD_C1,  M_BOO,   ModuleVersion(1,3,0))},
+  {PlatformVersion(P_FORD_C1,  M_EPS,   ModuleVersion(1,3,0))},
+  {PlatformVersion(P_FORD_CD4, M_BPEC,  ModuleVersion(2,6,0))},
+  {PlatformVersion(P_FORD_CD4, M_TPEC,  ModuleVersion(2,6,0))},
+  {PlatformVersion(P_FORD_CD4, M_STEER, ModuleVersion(2,6,0))},
+  {PlatformVersion(P_FORD_CD4, M_SHIFT, ModuleVersion(2,6,0))},
+  {PlatformVersion(P_FORD_CD5, M_BOO,   ModuleVersion(1,2,0))},
+  {PlatformVersion(P_FORD_CD5, M_TPEC,  ModuleVersion(1,2,0))},
+  {PlatformVersion(P_FORD_CD5, M_STEER, ModuleVersion(1,2,0))},
+  {PlatformVersion(P_FORD_GE1, M_TPEC,  ModuleVersion(1,0,0))},
+  {PlatformVersion(P_FORD_GE1, M_STEER, ModuleVersion(1,0,0))},
+  {PlatformVersion(P_FORD_GE1, M_SHIFT, ModuleVersion(1,0,0))},
+  {PlatformVersion(P_FORD_P5,  M_TPEC,  ModuleVersion(1,5,0))},
+  {PlatformVersion(P_FORD_P5,  M_STEER, ModuleVersion(1,5,0))},
+  {PlatformVersion(P_FORD_P5,  M_SHIFT, ModuleVersion(1,5,0))},
+  {PlatformVersion(P_FORD_P5,  M_ABS,   ModuleVersion(1,5,0))},
+  {PlatformVersion(P_FORD_P5,  M_BOO,   ModuleVersion(1,5,0))},
+  {PlatformVersion(P_FORD_T6,  M_TPEC,  ModuleVersion(0,3,0))},
+  {PlatformVersion(P_FORD_T6,  M_STEER, ModuleVersion(0,3,0))},
+  {PlatformVersion(P_FORD_T6,  M_SHIFT, ModuleVersion(0,3,0))},
+  {PlatformVersion(P_FORD_U6,  M_TPEC,  ModuleVersion(1,1,0))},
+  {PlatformVersion(P_FORD_U6,  M_STEER, ModuleVersion(1,1,0))},
+  {PlatformVersion(P_FORD_U6,  M_SHIFT, ModuleVersion(1,1,0))},
+  {PlatformVersion(P_FORD_U6,  M_ABS,   ModuleVersion(1,1,0))},
+  {PlatformVersion(P_FORD_U6,  M_BOO,   ModuleVersion(1,1,0))},
 });
 
 // Minimum firmware versions required for the timeout bit
@@ -161,12 +170,12 @@ DbwNode::DbwNode(ros::NodeHandle &node, ros::NodeHandle &priv_nh)
   joint_state_.velocity.resize(JOINT_COUNT);
   joint_state_.effort.resize(JOINT_COUNT);
   joint_state_.name.resize(JOINT_COUNT);
-  joint_state_.name[JOINT_FL] = "wheel_fl"; // Front Left
-  joint_state_.name[JOINT_FR] = "wheel_fr"; // Front Right
-  joint_state_.name[JOINT_RL] = "wheel_rl"; // Rear Left
-  joint_state_.name[JOINT_RR] = "wheel_rr"; // Rear Right
-  joint_state_.name[JOINT_SL] = "steer_fl";
-  joint_state_.name[JOINT_SR] = "steer_fr";
+  joint_state_.name[JOINT_FL] = "wheel_fl_joint"; // Front Left
+  joint_state_.name[JOINT_FR] = "wheel_fr_joint"; // Front Right
+  joint_state_.name[JOINT_RL] = "wheel_rl_joint"; // Rear Left
+  joint_state_.name[JOINT_RR] = "wheel_rr_joint"; // Rear Right
+  joint_state_.name[JOINT_SL] = "steer_fl_joint";
+  joint_state_.name[JOINT_SR] = "steer_fr_joint";
 
   // Setup Publishers
   pub_can_ = node.advertise<can_msgs::Frame>("can_tx", 10);
@@ -188,12 +197,17 @@ DbwNode::DbwNode(ros::NodeHandle &node, ros::NodeHandle &priv_nh)
   pub_gps_fix_ = node.advertise<sensor_msgs::NavSatFix>("gps/fix", 10);
   pub_gps_vel_ = node.advertise<geometry_msgs::TwistStamped>("gps/vel", 10);
   pub_gps_time_ = node.advertise<sensor_msgs::TimeReference>("gps/time", 10);
-  pub_joint_states_ = node.advertise<sensor_msgs::JointState>("joint_states", 10);
   pub_twist_ = node.advertise<geometry_msgs::TwistStamped>("twist", 10);
   pub_vin_ = node.advertise<std_msgs::String>("vin", 1, true);
   pub_sys_enable_ = node.advertise<std_msgs::Bool>("dbw_enabled", 1, true);
   publishDbwEnabled();
 
+  // Publish joint states if enabled
+  priv_nh.param("joint_states", enable_joint_states_, true);
+  if (enable_joint_states_) {
+    pub_joint_states_ = node.advertise<sensor_msgs::JointState>("joint_states", 10);
+  }
+
   // Setup Subscribers
   const ros::TransportHints NODELAY = ros::TransportHints().tcpNoDelay();
   sub_enable_ = node.subscribe("enable", 10, &DbwNode::recvEnable, this, NODELAY);
@@ -203,7 +217,8 @@ DbwNode::DbwNode(ros::NodeHandle &node, ros::NodeHandle &priv_nh)
   sub_throttle_ = node.subscribe("throttle_cmd", 1, &DbwNode::recvThrottleCmd, this, NODELAY);
   sub_steering_ = node.subscribe("steering_cmd", 1, &DbwNode::recvSteeringCmd, this, NODELAY);
   sub_gear_ = node.subscribe("gear_cmd", 1, &DbwNode::recvGearCmd, this, NODELAY);
-  sub_turn_signal_ = node.subscribe("turn_signal_cmd", 1, &DbwNode::recvTurnSignalCmd, this, NODELAY);
+  sub_turn_signal_ = node.subscribe("turn_signal_cmd", 1, &DbwNode::recvMiscCmd, this, NODELAY); // Backwards compatiblity
+  sub_misc_ = node.subscribe("misc_cmd", 1, &DbwNode::recvMiscCmd, this, NODELAY);
 
   // Setup Timer
   timer_ = node.createTimer(ros::Duration(1 / 20.0), &DbwNode::timerCallback, this);
@@ -236,7 +251,7 @@ void DbwNode::recvCAN(const can_msgs::Frame::ConstPtr& msg)
           faultWatchdog(ptr->FLTWDC, ptr->WDCSRC, ptr->WDCBRK);
           dbw_mkz_msgs::BrakeReport out;
           out.header.stamp = msg->header.stamp;
-          if (ptr->BTYPE == 0) {
+          if (ptr->BTYPE == 0 || firmware_.findModule(M_BPEC).valid()) {
             // Brake pedal emulator for hybrid electric vehicles
             out.pedal_input  = (float)ptr->PI / UINT16_MAX;
             out.pedal_cmd    = (float)ptr->PC / UINT16_MAX;
@@ -244,22 +259,20 @@ void DbwNode::recvCAN(const can_msgs::Frame::ConstPtr& msg)
             out.torque_input  = brakeTorqueFromPedal(out.pedal_input);
             out.torque_cmd    = brakeTorqueFromPedal(out.pedal_cmd);
             out.torque_output = brakeTorqueFromPedal(out.pedal_output);
-            out.decel_cmd = 0;
-            out.decel_output = 0;
-          } else {
+          } else if (ptr->BTYPE == 1 || firmware_.findModule(M_ABS).valid()) {
             // ACC/AEB braking for non-hybrid vehicles
-            out.pedal_input = 0;
-            out.pedal_cmd = 0;
-            out.pedal_output = 0;
             out.torque_input = ptr->PI;
-            out.torque_cmd = 0;
-            out.torque_output = 0;
             out.decel_cmd    = ptr->PC * 1e-3f;
             out.decel_output = ptr->PO * 1e-3f;
+          } else if (ptr->BTYPE == 2) {
+            // Brake pedal actuator for vehicles without brake-by-wire
+            out.torque_input = ptr->PI;
+            out.torque_cmd = ptr->PC;
+            out.torque_output = ptr->PO;
+          } else {
+            ROS_WARN_THROTTLE(5.0, "Unsupported brake report type: %u", ptr->BTYPE);
           }
-          out.boo_input  = ptr->BI ? true : false;
           out.boo_cmd    = ptr->BC ? true : false;
-          out.boo_output = ptr->BI || ptr->BC;
           out.enabled = ptr->ENABLED ? true : false;
           out.override = ptr->OVERRIDE ? true : false;
           out.driver = ptr->DRIVER ? true : false;
@@ -343,10 +356,10 @@ void DbwNode::recvCAN(const can_msgs::Frame::ConstPtr& msg)
           } else {
             out.steering_wheel_torque = (float)ptr->TORQUE * (float)0.0625;
           }
-          if (ptr->SPEED == 0xFFFF) {
+          if ((uint16_t)ptr->VEH_VEL == 0x8000) {
             out.speed = NAN;
           } else {
-            out.speed = (float)ptr->SPEED * (float)(0.01 / 3.6) * (float)speedSign();
+            out.speed = (float)ptr->VEH_VEL * (float)(0.01 / 3.6);
           }
           out.enabled = ptr->ENABLED ? true : false;
           out.override = ptr->OVERRIDE ? true : false;
@@ -367,7 +380,9 @@ void DbwNode::recvCAN(const can_msgs::Frame::ConstPtr& msg)
           twist.twist.linear.x = out.speed;
           twist.twist.angular.z = out.speed * tan(out.steering_wheel_angle / steering_ratio_) / acker_wheelbase_;
           pub_twist_.publish(twist);
-          publishJointStates(msg->header.stamp, NULL, &out);
+          if (enable_joint_states_) {
+            publishJointStates(msg->header.stamp, NULL, &out);
+          }
           if (ptr->FLTBUS1 || ptr->FLTBUS2 || ptr->FLTPWR) {
             ROS_WARN_THROTTLE(5.0, "Steering fault. FLT1: %s FLT2: %s FLTPWR: %s",
                 ptr->FLTBUS1 ? "true, " : "false,",
@@ -522,7 +537,9 @@ void DbwNode::recvCAN(const can_msgs::Frame::ConstPtr& msg)
             out.rear_right = (float)ptr->rear_right * 0.01f;
           }
           pub_wheel_speeds_.publish(out);
-          publishJointStates(msg->header.stamp, &out, NULL);
+          if (enable_joint_states_) {
+            publishJointStates(msg->header.stamp, &out, NULL);
+          }
         }
         break;
 
@@ -683,7 +700,13 @@ void DbwNode::recvCAN(const can_msgs::Frame::ConstPtr& msg)
           } else {
             out.engine_rpm = (float)ptr->engine_rpm * 0.25f;
           }
-          out.gear_num.num = ptr->gear_num;      
+          out.gear_num.num = ptr->gear_num;
+          out.ignition.value = ptr->ign_stat;
+          if ((uint16_t)ptr->batt_curr == 0xE000) {
+            out.batt_curr = NAN;
+          } else {
+            out.batt_curr = (float)ptr->batt_curr * 0.0625f;
+          }
           pub_throttle_info_.publish(out);
           if (ptr->aped_qf != dbw_mkz_msgs::QualityFactor::OK) {
             ROS_WARN_THROTTLE(5.0, "Throttle pedal limp-home: %u", ptr->aped_qf);
@@ -804,15 +827,21 @@ void DbwNode::recvCAN(const can_msgs::Frame::ConstPtr& msg)
               pub_vin_.publish(msg);
               ROS_INFO("Licensing: VIN: %s", vin_.c_str());
             }
-          } else if (ptr->mux == LIC_MUX_F0) {
-            const char * const NAME = "BASE"; // Base functionality
+          } else if ((LIC_MUX_F0 <= ptr->mux) && (ptr->mux <= LIC_MUX_F7)) {
+            constexpr std::array<const char*, 8> NAME = {"BASE","CONTROL","SENSORS","REMOTE","","","",""};
+            constexpr std::array<bool, 8> WARN = {true, true, true, false, true, true, true, true};
+            const size_t i = ptr->mux - LIC_MUX_F0;
+            const int id = module * NAME.size() + i;
+            const std::string name = strcmp(NAME[i], "") ? NAME[i] : std::string(1, '0' + i);
             if (ptr->license.enabled) {
-              ROS_INFO_ONCE_ID(module, "Licensing: %s feature '%s' enabled%s", str_m, NAME, ptr->license.trial ? " as a counted trial" : "");
+              ROS_INFO_ONCE_ID(id, "Licensing: %s feature '%s' enabled%s", str_m, name.c_str(), ptr->license.trial ? " as a counted trial" : "");
+            } else if (ptr->ready && !WARN[i]) {
+              ROS_INFO_ONCE_ID(id, "Licensing: %s feature '%s' not licensed.", str_m, name.c_str());
             } else if (ptr->ready) {
-              ROS_WARN_ONCE_ID(module, "Licensing: %s feature '%s' not licensed. Visit https://www.dataspeedinc.com/products/maintenance-subscription/ to request a license.", str_m, NAME);
+              ROS_WARN_ONCE_ID(id, "Licensing: %s feature '%s' not licensed. Visit https://www.dataspeedinc.com/products/maintenance-subscription/ to request a license.", str_m, name.c_str());
             }
-            if (ptr->ready && (module == M_STEER) && (ptr->license.trial || !ptr->license.enabled)) {
-              ROS_INFO_ONCE("Licensing: Feature '%s' trials used: %u, remaining: %u", NAME,
+            if (ptr->ready && (module == M_STEER) && (ptr->license.trial || (!ptr->license.enabled && WARN[i]))) {
+              ROS_INFO_ONCE("Licensing: Feature '%s' trials used: %u, remaining: %u", name.c_str(),
                             ptr->license.trials_used, ptr->license.trials_left);
             }
           }
@@ -844,18 +873,20 @@ void DbwNode::recvCAN(const can_msgs::Frame::ConstPtr& msg)
         break;
 
       case ID_BRAKE_CMD:
-        ROS_WARN_COND(warn_cmds_ && !((const MsgBrakeCmd*)msg->data.elems)->RES1,
+        ROS_WARN_COND(warn_cmds_ && !((const MsgBrakeCmd*)msg->data.elems)->RES1 && !((const MsgBrakeCmd*)msg->data.elems)->RES2,
                                   "DBW system: Another node on the CAN bus is commanding the vehicle!!! Subsystem: Brake. Id: 0x%03X", ID_BRAKE_CMD);
         break;
       case ID_THROTTLE_CMD:
-        ROS_WARN_COND(warn_cmds_ && !((const MsgThrottleCmd*)msg->data.elems)->RES1,
+        ROS_WARN_COND(warn_cmds_ && !((const MsgThrottleCmd*)msg->data.elems)->RES1 && !((const MsgThrottleCmd*)msg->data.elems)->RES2,
                                   "DBW system: Another node on the CAN bus is commanding the vehicle!!! Subsystem: Throttle. Id: 0x%03X", ID_THROTTLE_CMD);
         break;
       case ID_STEERING_CMD:
-        ROS_WARN_COND(warn_cmds_, "DBW system: Another node on the CAN bus is commanding the vehicle!!! Subsystem: Steering. Id: 0x%03X", ID_STEERING_CMD);
+        ROS_WARN_COND(warn_cmds_ && !((const MsgSteeringCmd*)msg->data.elems)->RES1 && !((const MsgSteeringCmd*)msg->data.elems)->RES2,
+                                  "DBW system: Another node on the CAN bus is commanding the vehicle!!! Subsystem: Steering. Id: 0x%03X", ID_STEERING_CMD);
         break;
       case ID_GEAR_CMD:
-        ROS_WARN_COND(warn_cmds_, "DBW system: Another node on the CAN bus is commanding the vehicle!!! Subsystem: Shifting. Id: 0x%03X", ID_GEAR_CMD);
+        ROS_WARN_COND(warn_cmds_ && !((const MsgGearCmd*)msg->data.elems)->RES1 && !((const MsgGearCmd*)msg->data.elems)->RES2,
+                                  "DBW system: Another node on the CAN bus is commanding the vehicle!!! Subsystem: Shifting. Id: 0x%03X", ID_GEAR_CMD);
         break;
       case ID_MISC_CMD:
         ROS_WARN_COND(warn_cmds_, "DBW system: Another node on the CAN bus is commanding the vehicle!!! Subsystem: Turn Signals. Id: 0x%03X", ID_MISC_CMD);
@@ -996,96 +1027,94 @@ void DbwNode::recvBrakeCmd(const dbw_mkz_msgs::BrakeCmd::ConstPtr& msg)
   out.dlc = sizeof(MsgBrakeCmd);
   MsgBrakeCmd *ptr = (MsgBrakeCmd*)out.data.elems;
   memset(ptr, 0x00, sizeof(*ptr));
-  if (enabled()) {
-    bool fwd_abs = firmware_.findModule(M_ABS).valid(); // Does the ABS braking module exist?
-    bool fwd_bpe = firmware_.findPlatform(M_BPEC) >= FIRMWARE_CMDTYPE; // Minimum required BPEC firmware version
-    bool fwd = !pedal_luts_; // Forward command type, or apply pedal LUTs locally
-    fwd |= fwd_abs; // The local pedal LUTs are for the BPEC module, the ABS module requires forwarding
-    fwd &= fwd_bpe; // Only modern BPEC firmware supports forwarding the command type
-    switch (msg->pedal_cmd_type) {
-      case dbw_mkz_msgs::BrakeCmd::CMD_NONE:
-        break;
-      case dbw_mkz_msgs::BrakeCmd::CMD_PEDAL:
+  bool fwd_abs = firmware_.findModule(M_ABS).valid(); // Does the ABS braking module exist?
+  bool fwd_bpe = firmware_.findPlatform(M_BPEC) >= FIRMWARE_CMDTYPE; // Minimum required BPEC firmware version
+  bool fwd = !pedal_luts_; // Forward command type, or apply pedal LUTs locally
+  fwd |= fwd_abs; // The local pedal LUTs are for the BPEC module, the ABS module requires forwarding
+  fwd &= fwd_bpe; // Only modern BPEC firmware supports forwarding the command type
+  switch (msg->pedal_cmd_type) {
+    case dbw_mkz_msgs::BrakeCmd::CMD_NONE:
+      break;
+    case dbw_mkz_msgs::BrakeCmd::CMD_PEDAL:
+      ptr->CMD_TYPE = dbw_mkz_msgs::BrakeCmd::CMD_PEDAL;
+      ptr->PCMD = std::clamp<float>(msg->pedal_cmd * UINT16_MAX, 0, UINT16_MAX);
+      if (!firmware_.findModule(M_BPEC).valid() && firmware_.findModule(M_ABS).valid()) {
+        ROS_WARN_THROTTLE(1.0, "Module ABS does not support brake command type PEDAL");
+      }
+      break;
+    case dbw_mkz_msgs::BrakeCmd::CMD_PERCENT:
+      if (fwd) {
+        ptr->CMD_TYPE = dbw_mkz_msgs::BrakeCmd::CMD_PERCENT;
+        ptr->PCMD = std::clamp<float>(msg->pedal_cmd * UINT16_MAX, 0, UINT16_MAX);
+      } else {
         ptr->CMD_TYPE = dbw_mkz_msgs::BrakeCmd::CMD_PEDAL;
-        ptr->PCMD = std::max((float)0.0, std::min((float)UINT16_MAX, msg->pedal_cmd * UINT16_MAX));
-        if (!firmware_.findModule(M_BPEC).valid() && firmware_.findModule(M_ABS).valid()) {
-          ROS_WARN_THROTTLE(1.0, "Module ABS does not support brake command type PEDAL");
-        }
-        break;
-      case dbw_mkz_msgs::BrakeCmd::CMD_PERCENT:
-        if (fwd) {
-          ptr->CMD_TYPE = dbw_mkz_msgs::BrakeCmd::CMD_PERCENT;
-          ptr->PCMD = std::max((float)0.0, std::min((float)UINT16_MAX, msg->pedal_cmd * UINT16_MAX));
-        } else {
-          ptr->CMD_TYPE = dbw_mkz_msgs::BrakeCmd::CMD_PEDAL;
-          ptr->PCMD = std::max((float)0.0, std::min((float)UINT16_MAX, brakePedalFromPercent(msg->pedal_cmd) * UINT16_MAX));
-        }
-        break;
-      case dbw_mkz_msgs::BrakeCmd::CMD_TORQUE:
-        if (fwd) {
-          ptr->CMD_TYPE = dbw_mkz_msgs::BrakeCmd::CMD_TORQUE;
-          ptr->PCMD = std::max((float)0.0, std::min((float)UINT16_MAX, msg->pedal_cmd));
-        } else {
-          ptr->CMD_TYPE = dbw_mkz_msgs::BrakeCmd::CMD_PEDAL;
-          ptr->PCMD = std::max((float)0.0, std::min((float)UINT16_MAX, brakePedalFromTorque(msg->pedal_cmd) * UINT16_MAX));
-        }
-        if (!firmware_.findModule(M_BPEC).valid() && firmware_.findModule(M_ABS).valid()) {
-          ROS_WARN_THROTTLE(1.0, "Module ABS does not support brake command type TORQUE");
-        }
-        break;
-      case dbw_mkz_msgs::BrakeCmd::CMD_TORQUE_RQ:
-        if (fwd_abs || fwd_bpe) {
-          // CMD_TORQUE_RQ must be forwarded, there is no local implementation
-          fwd = true;
-          ptr->CMD_TYPE = dbw_mkz_msgs::BrakeCmd::CMD_TORQUE_RQ;
-          ptr->PCMD = std::max((float)0.0, std::min((float)UINT16_MAX, msg->pedal_cmd));
-        } else if (fwd) {
-          // Fallback to forwarded CMD_TORQUE
-          ptr->CMD_TYPE = dbw_mkz_msgs::BrakeCmd::CMD_TORQUE;
-          ptr->PCMD = std::max((float)0.0, std::min((float)UINT16_MAX, msg->pedal_cmd));
-        } else {
-          // Fallback to local CMD_TORQUE
-          ptr->CMD_TYPE = dbw_mkz_msgs::BrakeCmd::CMD_PEDAL;
-          ptr->PCMD = std::max((float)0.0, std::min((float)UINT16_MAX, brakePedalFromTorque(msg->pedal_cmd) * UINT16_MAX));
-        }
-        if (!firmware_.findModule(M_BPEC).valid() && firmware_.findModule(M_ABS).valid()) {
-          ROS_WARN_THROTTLE(1.0, "Module ABS does not support brake command type TORQUE_RQ");
-        }
-        break;
-      case dbw_mkz_msgs::BrakeCmd::CMD_DECEL:
-        // CMD_DECEL must be forwarded, there is no local implementation
-        ptr->CMD_TYPE = dbw_mkz_msgs::BrakeCmd::CMD_DECEL;
-        ptr->PCMD = std::max((float)0.0, std::min((float)10e3, msg->pedal_cmd * 1e3f));
-        if (!firmware_.findModule(M_ABS).valid() && firmware_.findModule(M_BPEC).valid()) {
-          ROS_WARN_THROTTLE(1.0, "Module BPEC does not support brake command type DECEL");
-        }
-        break;
-      default:
-        ROS_WARN("Unknown brake command type: %u", msg->pedal_cmd_type);
-        break;
-    }
+        ptr->PCMD = std::clamp<float>(brakePedalFromPercent(msg->pedal_cmd) * UINT16_MAX, 0, UINT16_MAX);
+      }
+      break;
+    case dbw_mkz_msgs::BrakeCmd::CMD_TORQUE:
+      if (fwd) {
+        ptr->CMD_TYPE = dbw_mkz_msgs::BrakeCmd::CMD_TORQUE;
+        ptr->PCMD = std::clamp<float>(msg->pedal_cmd, 0, UINT16_MAX);
+      } else {
+        ptr->CMD_TYPE = dbw_mkz_msgs::BrakeCmd::CMD_PEDAL;
+        ptr->PCMD = std::clamp<float>(brakePedalFromTorque(msg->pedal_cmd) * UINT16_MAX, 0, UINT16_MAX);
+      }
+      if (!firmware_.findModule(M_BPEC).valid() && firmware_.findModule(M_ABS).valid()) {
+        ROS_WARN_THROTTLE(1.0, "Module ABS does not support brake command type TORQUE");
+      }
+      break;
+    case dbw_mkz_msgs::BrakeCmd::CMD_TORQUE_RQ:
+      if (fwd_abs || fwd_bpe) {
+        // CMD_TORQUE_RQ must be forwarded, there is no local implementation
+        fwd = true;
+        ptr->CMD_TYPE = dbw_mkz_msgs::BrakeCmd::CMD_TORQUE_RQ;
+        ptr->PCMD = std::clamp<float>(msg->pedal_cmd, 0, UINT16_MAX);
+      } else if (fwd) {
+        // Fallback to forwarded CMD_TORQUE
+        ptr->CMD_TYPE = dbw_mkz_msgs::BrakeCmd::CMD_TORQUE;
+        ptr->PCMD = std::clamp<float>(msg->pedal_cmd, 0, UINT16_MAX);
+      } else {
+        // Fallback to local CMD_TORQUE
+        ptr->CMD_TYPE = dbw_mkz_msgs::BrakeCmd::CMD_PEDAL;
+        ptr->PCMD = std::clamp<float>(brakePedalFromTorque(msg->pedal_cmd) * UINT16_MAX, 0, UINT16_MAX);
+      }
+      if (!firmware_.findModule(M_BPEC).valid() && firmware_.findModule(M_ABS).valid()) {
+        ROS_WARN_THROTTLE(1.0, "Module ABS does not support brake command type TORQUE_RQ");
+      }
+      break;
+    case dbw_mkz_msgs::BrakeCmd::CMD_DECEL:
+      // CMD_DECEL must be forwarded, there is no local implementation
+      ptr->CMD_TYPE = dbw_mkz_msgs::BrakeCmd::CMD_DECEL;
+      ptr->PCMD = std::clamp<float>(msg->pedal_cmd * 1e3f, 0, 10e3);
+      if (!firmware_.findModule(M_ABS).valid() && firmware_.findModule(M_BPEC).valid()) {
+        ROS_WARN_THROTTLE(1.0, "Module BPEC does not support brake command type DECEL");
+      }
+      break;
+    default:
+      ROS_WARN("Unknown brake command type: %u", msg->pedal_cmd_type);
+      break;
+  }
 #if 1 // Manually implement auto BOO control (brake lights) for legacy firmware
-    ptr->ABOO = 1;
-    const PlatformVersion firmware_bpec = firmware_.findPlatform(M_BPEC);
-    if (firmware_bpec.v.valid() && (firmware_bpec < FIRMWARE_CMDTYPE)) {
-      const uint16_t BOO_THRESH_LO = 0.20 * UINT16_MAX;
-      const uint16_t BOO_THRESH_HI = 0.22 * UINT16_MAX;
-      static bool boo_status_ = false;
-      if (boo_status_) {
-        ptr->BCMD = 1;
-      }
-      if (!boo_status_ && (ptr->PCMD > BOO_THRESH_HI)) {
-        ptr->BCMD = 1;
-        boo_status_ = true;
-      } else if (boo_status_ && (ptr->PCMD < BOO_THRESH_LO)) {
-        ptr->BCMD = 0;
-        boo_status_ = false;
-      }
+  ptr->ABOO = 1;
+  const PlatformVersion firmware_bpec = firmware_.findPlatform(M_BPEC);
+  if (firmware_bpec.v.valid() && (firmware_bpec < FIRMWARE_CMDTYPE)) {
+    const uint16_t BOO_THRESH_LO = 0.20 * UINT16_MAX;
+    const uint16_t BOO_THRESH_HI = 0.22 * UINT16_MAX;
+    static bool boo_status_ = false;
+    if (boo_status_) {
+      ptr->BCMD = 1;
     }
+    if (!boo_status_ && (ptr->PCMD > BOO_THRESH_HI)) {
+      ptr->BCMD = 1;
+      boo_status_ = true;
+    } else if (boo_status_ && (ptr->PCMD < BOO_THRESH_LO)) {
+      ptr->BCMD = 0;
+      boo_status_ = false;
+    }
+  }
 #endif
-    if (msg->enable) {
-      ptr->EN = 1;
-    }
+  if (enabled() && msg->enable) {
+    ptr->EN = 1;
   }
   if (clear() || msg->clear) {
     ptr->CLEAR = 1;
@@ -1105,34 +1134,32 @@ void DbwNode::recvThrottleCmd(const dbw_mkz_msgs::ThrottleCmd::ConstPtr& msg)
   out.dlc = sizeof(MsgThrottleCmd);
   MsgThrottleCmd *ptr = (MsgThrottleCmd*)out.data.elems;
   memset(ptr, 0x00, sizeof(*ptr));
-  if (enabled()) {
-    bool fwd = !pedal_luts_; // Forward command type, or apply pedal LUTs locally
-    fwd &= firmware_.findPlatform(M_TPEC) >= FIRMWARE_CMDTYPE; // Minimum required firmware version
-    float cmd = 0.0;
-    switch (msg->pedal_cmd_type) {
-      case dbw_mkz_msgs::ThrottleCmd::CMD_NONE:
-        break;
-      case dbw_mkz_msgs::ThrottleCmd::CMD_PEDAL:
-        ptr->CMD_TYPE = dbw_mkz_msgs::ThrottleCmd::CMD_PEDAL;
+  bool fwd = !pedal_luts_; // Forward command type, or apply pedal LUTs locally
+  fwd &= firmware_.findPlatform(M_TPEC) >= FIRMWARE_CMDTYPE; // Minimum required firmware version
+  float cmd = 0.0;
+  switch (msg->pedal_cmd_type) {
+    case dbw_mkz_msgs::ThrottleCmd::CMD_NONE:
+      break;
+    case dbw_mkz_msgs::ThrottleCmd::CMD_PEDAL:
+      ptr->CMD_TYPE = dbw_mkz_msgs::ThrottleCmd::CMD_PEDAL;
+      cmd = msg->pedal_cmd;
+      break;
+    case dbw_mkz_msgs::ThrottleCmd::CMD_PERCENT:
+      if (fwd) {
+        ptr->CMD_TYPE = dbw_mkz_msgs::ThrottleCmd::CMD_PERCENT;
         cmd = msg->pedal_cmd;
-        break;
-      case dbw_mkz_msgs::ThrottleCmd::CMD_PERCENT:
-        if (fwd) {
-          ptr->CMD_TYPE = dbw_mkz_msgs::ThrottleCmd::CMD_PERCENT;
-          cmd = msg->pedal_cmd;
-        } else {
-          ptr->CMD_TYPE = dbw_mkz_msgs::ThrottleCmd::CMD_PEDAL;
-          cmd = throttlePedalFromPercent(msg->pedal_cmd);
-        }
-        break;
-      default:
-        ROS_WARN("Unknown throttle command type: %u", msg->pedal_cmd_type);
-        break;
-    }
-    ptr->PCMD = std::max((float)0.0, std::min((float)UINT16_MAX, cmd * UINT16_MAX));
-    if (msg->enable) {
-      ptr->EN = 1;
-    }
+      } else {
+        ptr->CMD_TYPE = dbw_mkz_msgs::ThrottleCmd::CMD_PEDAL;
+        cmd = throttlePedalFromPercent(msg->pedal_cmd);
+      }
+      break;
+    default:
+      ROS_WARN("Unknown throttle command type: %u", msg->pedal_cmd_type);
+      break;
+  }
+  ptr->PCMD = std::clamp<float>(cmd * UINT16_MAX, 0, UINT16_MAX);
+  if (enabled() && msg->enable) {
+    ptr->EN = 1;
   }
   if (clear() || msg->clear) {
     ptr->CLEAR = 1;
@@ -1152,33 +1179,31 @@ void DbwNode::recvSteeringCmd(const dbw_mkz_msgs::SteeringCmd::ConstPtr& msg)
   out.dlc = sizeof(MsgSteeringCmd);
   MsgSteeringCmd *ptr = (MsgSteeringCmd*)out.data.elems;
   memset(ptr, 0x00, sizeof(*ptr));
-  if (enabled()) {
-    switch (msg->cmd_type) {
-      case dbw_mkz_msgs::SteeringCmd::CMD_ANGLE:
-        ptr->SCMD = std::max((float)-INT16_MAX, std::min((float)INT16_MAX, (float)(msg->steering_wheel_angle_cmd * (180 / M_PI * 10))));
-        if (fabsf(msg->steering_wheel_angle_velocity) > 0) {
-          if (firmware_.findModule(M_EPS).valid() || (firmware_.findPlatform(M_STEER) >= FIRMWARE_HIGH_RATE_LIMIT)) {
-            ptr->SVEL = std::max((float)1, std::min((float)254, (float)roundf(fabsf(msg->steering_wheel_angle_velocity) * 180 / M_PI / 4)));
-          } else {
-            ptr->SVEL = std::max((float)1, std::min((float)254, (float)roundf(fabsf(msg->steering_wheel_angle_velocity) * 180 / M_PI / 2)));
-          }
+  switch (msg->cmd_type) {
+    case dbw_mkz_msgs::SteeringCmd::CMD_ANGLE:
+      ptr->SCMD = std::clamp<float>(msg->steering_wheel_angle_cmd * (float)(180 / M_PI * 10), -INT16_MAX, INT16_MAX);
+      if (fabsf(msg->steering_wheel_angle_velocity) > 0) {
+        if (firmware_.findModule(M_EPS).valid() || (firmware_.findPlatform(M_STEER) >= FIRMWARE_HIGH_RATE_LIMIT)) {
+          ptr->SVEL = std::clamp<float>(roundf(fabsf(msg->steering_wheel_angle_velocity) * (float)(180 / M_PI / 4)), 1, 254);
+        } else {
+          ptr->SVEL = std::clamp<float>(roundf(fabsf(msg->steering_wheel_angle_velocity) * (float)(180 / M_PI / 2)), 1, 254);
         }
-        ptr->CMD_TYPE = dbw_mkz_msgs::SteeringCmd::CMD_ANGLE;
-        break;
-      case dbw_mkz_msgs::SteeringCmd::CMD_TORQUE:
-        ptr->SCMD = std::max((float)-INT16_MAX, std::min((float)INT16_MAX, (float)(msg->steering_wheel_torque_cmd * 128)));
-        ptr->CMD_TYPE = dbw_mkz_msgs::SteeringCmd::CMD_TORQUE;
-        if (!firmware_.findModule(M_EPS).valid() && firmware_.findModule(M_STEER).valid()) {
-          ROS_WARN_THROTTLE(1.0, "Module STEER does not support steering command type TORQUE");
-        }
-        break;
-      default:
-        ROS_WARN("Unknown steering command type: %u", msg->cmd_type);
-        break;
-    }
-    if (msg->enable) {
-      ptr->EN = 1;
-    }
+      }
+      ptr->CMD_TYPE = dbw_mkz_msgs::SteeringCmd::CMD_ANGLE;
+      break;
+    case dbw_mkz_msgs::SteeringCmd::CMD_TORQUE:
+      ptr->SCMD = std::clamp<float>(msg->steering_wheel_torque_cmd * 128, -INT16_MAX, INT16_MAX);
+      ptr->CMD_TYPE = dbw_mkz_msgs::SteeringCmd::CMD_TORQUE;
+      if (!firmware_.findModule(M_EPS).valid() && firmware_.findModule(M_STEER).valid()) {
+        ROS_WARN_THROTTLE(1.0, "Module STEER does not support steering command type TORQUE");
+      }
+      break;
+    default:
+      ROS_WARN("Unknown steering command type: %u", msg->cmd_type);
+      break;
+  }
+  if (enabled() && msg->enable) {
+    ptr->EN = 1;
   }
   if (clear() || msg->clear) {
     ptr->CLEAR = 1;
@@ -1189,6 +1214,9 @@ void DbwNode::recvSteeringCmd(const dbw_mkz_msgs::SteeringCmd::ConstPtr& msg)
   if (msg->quiet) {
     ptr->QUIET = 1;
   }
+  if (msg->alert) {
+    ptr->ALERT = 1;
+  }
   ptr->COUNT = msg->count;
   pub_can_.publish(out);
 }
@@ -1210,29 +1238,29 @@ void DbwNode::recvGearCmd(const dbw_mkz_msgs::GearCmd::ConstPtr& msg)
   pub_can_.publish(out);
 }
 
-void DbwNode::recvTurnSignalCmd(const dbw_mkz_msgs::TurnSignalCmd::ConstPtr& msg)
+void DbwNode::recvMiscCmd(const dbw_mkz_msgs::MiscCmd::ConstPtr& msg)
 {
   can_msgs::Frame out;
   out.id = ID_MISC_CMD;
   out.is_extended = false;
-  out.dlc = sizeof(MsgTurnSignalCmd);
-  MsgTurnSignalCmd *ptr = (MsgTurnSignalCmd*)out.data.elems;
+  out.dlc = sizeof(MsgMiscCmd);
+  MsgMiscCmd *ptr = (MsgMiscCmd*)out.data.elems;
   memset(ptr, 0x00, sizeof(*ptr));
   if (enabled()) {
     ptr->TRNCMD = msg->cmd.value;
+    ptr->PBRKCMD = msg->pbrk.cmd;
   }
   pub_can_.publish(out);
 }
 
-bool DbwNode::publishDbwEnabled()
+bool DbwNode::publishDbwEnabled(bool force)
 {
-  bool change = false;
   bool en = enabled();
-  if (prev_enable_ != en) {
+  bool change = prev_enable_ != en;
+  if (change || force) {
     std_msgs::Bool msg;
     msg.data = en;
     pub_sys_enable_.publish(msg);
-    change = true;
   }
   prev_enable_ = en;
   return change;
@@ -1240,6 +1268,12 @@ bool DbwNode::publishDbwEnabled()
 
 void DbwNode::timerCallback(const ros::TimerEvent& event)
 {
+  // Publish status periodically, in addition to latched and on change
+  if (publishDbwEnabled(true)) {
+    ROS_WARN("DBW system enable status changed unexpectedly");
+  }
+
+  // Clear override statuses if necessary
   if (clear()) {
     can_msgs::Frame out;
     out.is_extended = false;
